class FortmaticDjangoException(Exception):
    pass


class UnsupportedAuthMode(FortmaticDjangoException):
    pass


class PublicAddressDoesNotExist(FortmaticDjangoException):
    pass


class UserEmailMissmatch(FortmaticDjangoException):
    pass


class UnableToLoadUserFromIdentityToken(FortmaticDjangoException):
    pass


class InvalidIdentityToken(FortmaticDjangoException):
    pass


class InvalidSignerAddress(FortmaticDjangoException):
    pass


class InvalidChallengeMessage(FortmaticDjangoException):
    pass



import re

AUTHORIZATION_PATTERN = r'Bearer (?P<token>.+)'


def null_safe(value):
    if value == 'null':
        return None

    return value


def parse_authorization_header(request):
    header_value = request.META.get('HTTP_AUTHORIZATION', '')

    m = re.match(AUTHORIZATION_PATTERN, header_value)

    if m is None:
        return None

    return null_safe(m.group('token'))



import json

import requests
from retrying import retry

from fortmatic.api_resources.auth.resource_uris import challenge_invalidate_v1
from fortmatic.api_resources.util.exceptions import APIError
from fortmatic.api_resources.util.request_helpers import get_call_uri
from fortmatic.api_resources.util.request_helpers import get_secret_api_key_headers
from fortmatic.api_resources.util.request_helpers import retry_on_request_exceptions


@retry(
    retry_on_exception=retry_on_request_exceptions,
    wait_exponential_multiplier=100,
    wait_exponential_max=2000,
    stop_max_attempt_number=3,
)
def invalidate_token_v1(public_address):
    resp = requests.post(
        get_call_uri(challenge_invalidate_v1),
        data=json.dumps({'public_address': public_address}),
        headers=get_secret_api_key_headers(),
    )

    data = resp.json()
    if resp.status_code >= 400:
        raise APIError(message=data['message'], error_code=data['error_code'])



import base64
import json

from eth_account.messages import encode_structured_data
from web3.auto import w3

from fortmatic.api_resources.auth.challenge_message import get_challenge_message_v1
from fortmatic.api_resources.auth.invalidate_token import invalidate_token_v1
from fortmatic.django.exceptions import InvalidChallengeMessage
from fortmatic.django.exceptions import InvalidIdentityToken
from fortmatic.django.exceptions import InvalidSignerAddress
from fortmatic.utils.http import parse_authorization_header


def get_identity_token_from_header(request):
    return parse_authorization_header(request)


class PhantomAuth:

    def __init__(self):
        pass

    @staticmethod
    def decode_token(identity_token):
        """
        Args:
            identity_token (base64.str): Base64 encoded string.

        Raises:
            InvalidIdentityToken: If token is malformed.
        """
        try:
            decoded_dat = json.loads(
                base64.urlsafe_b64decode(
                    identity_token,
                ).decode('utf-8'),
            )
        except Exception:
            raise InvalidIdentityToken()

        proof = decoded_dat[0]
        claim = decoded_dat[1]
        return proof, claim

    def _check_integratity_and_authenticity_of_the_token(
        self,
        identity_token,
    ):
        """
        Args:
            identity_token (base64.str): Base64 encoded string.

        Raises:
            InvalidSignerAddress: If unable to match recovered address with
                signer address.
        """
        proof, claim = self.decode_token(identity_token)
        signable_message = encode_structured_data(claim)
        recovered_address = w3.eth.account.recover_message(
            signable_message,
            signature=proof,
        )

        if recovered_address != claim['message']['addr']:
            raise InvalidSignerAddress()

    def _check_validity_of_the_token(self, identity_token):
        """
        Args:
            identity_token (base64.str): Base64 encoded string.

        Raises:
            InvalidChallengeMessage: If challenge message in ``identity_token``
                does not match API provided challenge message.
        """
        _, claim = self.decode_token(identity_token)
        if claim['message']['chmsg'] != self.get_challenge_message(identity_token):
            raise InvalidChallengeMessage()

    def validate_token(
        self,
        identity_token,
        skip_checking_challenge_message=False,
    ):
        """
        Args:
            identity_token (base64.str): Base64 encoded string.

        Returns:
            None.
        """
        self._check_integratity_and_authenticity_of_the_token(identity_token)
        self._check_validity_of_the_token(identity_token)

    def get_public_address(self, identity_token):
        """
        Args:
            identity_token (base64.str): Base64 encoded string.

        Raises:
            InvalidIdentityToken: If token is malformed.
        """
        self._check_integratity_and_authenticity_of_the_token(identity_token)

        _, claim = self.decode_token(identity_token)
        try:
            return claim['message']['addr']
        except KeyError:
            raise InvalidIdentityToken()

    def get_challenge_message(self, identity_token):
        return get_challenge_message_v1(
            self.get_public_address(identity_token),
        )

    def invalidate_token(self, identity_token):
        invalidate_token_v1(self.get_public_address(identity_token))

    def invalidate_token_by_public_address(self, public_address):
        invalidate_token_v1(public_address)



from requests.exceptions import RequestException

import fortmatic
from fortmatic.api_resources.config import RESOURCE_SERVER
from fortmatic.api_resources.config import SECRET_API_KEY_HEADER


def get_call_uri(resource):
    return RESOURCE_SERVER + resource


def get_secret_api_key_headers():
    if fortmatic.secret_key is None:
        raise Exception(
            'Please specify your account\'s secret key. You can get a secret '
            'key from Fortmatic Dashboard (https://dashboard.fortmatic.com). '
            'And in your application you can do: \n'
            '\t"import fortmatic; fortmatic.secret_key = \'sk_live_XYZ\'"',
        )

    return {SECRET_API_KEY_HEADER: fortmatic.secret_key}


def retry_on_request_exceptions(exception):
    return isinstance(exception, RequestException)



from django.contrib.auth import logout as django_logout

from fortmatic.django.config import FORTMATIC_IDENTITY_KEY
from fortmatic.utils.auth import PhantomAuth


def logout(request):
    identity_token = request.session.get(FORTMATIC_IDENTITY_KEY, None)
    if identity_token:
        PhantomAuth().invalidate_token(identity_token)

    django_logout(request)



from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend

from fortmatic.django.config import FORTMATIC_IDENTITY_KEY
from fortmatic.django.config import FortmaticAuthBackenMode
from fortmatic.django.exceptions import InvalidChallengeMessage
from fortmatic.django.exceptions import InvalidIdentityToken
from fortmatic.django.exceptions import InvalidSignerAddress
from fortmatic.django.exceptions import PublicAddressDoesNotExist
from fortmatic.django.exceptions import UnsupportedAuthMode
from fortmatic.django.exceptions import UserEmailMissmatch
from fortmatic.utils.auth import get_identity_token_from_header
from fortmatic.utils.auth import PhantomAuth


user_model = get_user_model()


class FortmaticAuthBackend(ModelBackend):

    @staticmethod
    def _load_user_from_email(email):
        try:
            return user_model.objects.get(email=email)
        except user_model.DoesNotExist:
            return None

    @staticmethod
    def _persist_data_in_session(request, identity_token):
        request.session[FORTMATIC_IDENTITY_KEY] = identity_token
        request.session.modified = True

    @staticmethod
    def _validate_identity_token_and_load_user(identity_token, email):
        try:
            PhantomAuth().validate_token(identity_token)
        except (
            InvalidIdentityToken,
            InvalidSignerAddress,
            InvalidChallengeMessage,
        ):
            return None

        public_address = PhantomAuth().get_public_address(identity_token)
        try:
            user = user_model.get_by_public_address(public_address)
        except user_model.DoesNotExist:
            raise PublicAddressDoesNotExist()

        if user.email != email:
            raise UserEmailMissmatch()

        return user

    def user_can_authenticate(self, user):
        if user is None:
            return False

        return super().user_can_authenticate(user)

    def _update_user_with_public_address(self, user, public_address):
        if self.user_can_authenticate(user):
            user.update_user_with_public_address(
                user_id=None,
                public_address=public_address,
                user_obj=user,
            )

    def _handle_phantom_auth(self, request, email):
        identity_token = get_identity_token_from_header(request)
        if identity_token is None:
            return None

        try:
            user = self._validate_identity_token_and_load_user(
                identity_token,
                email,
            )
        except PublicAddressDoesNotExist:
            user = self._load_user_from_email(email)
            if user is None:
                PhantomAuth().invalidate_token(identity_token)
                return None

            self._update_user_with_public_address(
                user,
                PhantomAuth().get_public_address(identity_token),
            )
        except UserEmailMissmatch:
            PhantomAuth().invalidate_token(identity_token)
            return None

        if self.user_can_authenticate(user):
            self._persist_data_in_session(request, identity_token)
            return user

        return None

    def authenticate(
        self,
        request,
        email=None,
        mode=FortmaticAuthBackenMode.PHANTOM,
    ):
        email = user_model.objects.normalize_email(email)

        if mode == FortmaticAuthBackenMode.PHANTOM:
            return self._handle_phantom_auth(request, email)
        else:
            raise UnsupportedAuthMode()



class APIError(Exception):
    message = None
    error_code = None

    def __init__(self, message=None, error_code=None):
        if message is not None:
            self.message = message

        if error_code is not None:
            self.error_code = error_code



from setuptools import setup
from setuptools import find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='fortmatic',
    version='0.0.3',
    description='Fortmatic Python Library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Fortmatic',
    author_email='support@fortmatic.com',
    url='https://fortmatic.com',
    license='MIT',
    keywords='fortmatic python sdk',
    packages=find_packages(
        exclude=[
            'tests',
            'tests.*',
            'examples',
            'examples.*',
            'testing',
            'testing.*',
        ],
    ),
    zip_safe=False,
    install_requires=[
        'requests == 2.22',
        'web3 == 5.4.0',
        'retrying == 1.3.3',
    ],
    python_requires='>=3.6',
    project_urls={
        'Website': 'https://fortmatic.com',
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



FORTMATIC_IDENTITY_KEY = '_fortmatic_identity_token'
FORTMATIC_AUTH_BACKEND = 'fortmatic.django.auth.backends.FortmaticAuthBackend'


class FortmaticAuthBackenMode:

    DJANGO_DEFAULT_AUTH = 0
    PHANTOM = 1

    ALLOWED_MODES = frozenset([DJANGO_DEFAULT_AUTH, PHANTOM])



from django.contrib.auth import get_user_model
from django.contrib.auth import login
from django.contrib.auth import logout
from django.contrib.auth.models import AnonymousUser
from django.utils.deprecation import MiddlewareMixin

from fortmatic.django.config import FORTMATIC_AUTH_BACKEND
from fortmatic.django.config import FORTMATIC_IDENTITY_KEY
from fortmatic.django.exceptions import InvalidChallengeMessage
from fortmatic.django.exceptions import InvalidIdentityToken
from fortmatic.django.exceptions import InvalidSignerAddress
from fortmatic.django.exceptions import UnableToLoadUserFromIdentityToken
from fortmatic.utils.auth import get_identity_token_from_header
from fortmatic.utils.auth import PhantomAuth


user_model = get_user_model()


class FortmaticAuthMiddleware(MiddlewareMixin):

    @staticmethod
    def _persist_data_in_session(request, identity_token):
        request.session[FORTMATIC_IDENTITY_KEY] = identity_token
        request.session.modified = True

    @staticmethod
    def _load_identity_token_from_session(request):
        return request.session.get(FORTMATIC_IDENTITY_KEY, None)

    @staticmethod
    def _load_identity_token_from_header(request):
        return get_identity_token_from_header(request)

    @staticmethod
    def _is_request_related_to_fortmatic(identity_token):
        return identity_token is not None

    @staticmethod
    def _try_loading_user_from_identity_token(identity_token):
        public_address = PhantomAuth().get_public_address(identity_token)

        try:
            return user_model.get_by_public_address(public_address)
        except user_model.DoesNotExist:
            return AnonymousUser()

    @staticmethod
    def can_user_be_logged_in(user):
        is_active = getattr(user, 'is_active', None)
        return is_active or is_active is None

    def _attempt_handling_anonymous_user(self, request, identity_token):
        user = self._try_loading_user_from_identity_token(identity_token)

        if not user.is_anonymous and self.can_user_be_logged_in(user):
            # Log the user in to rehydrate the session.
            login(request, user, backend=FORTMATIC_AUTH_BACKEND)
        else:
            raise UnableToLoadUserFromIdentityToken()

    def _load_identity_token_from_sources(self, request):
        # The identity token from header take precedence over the one in session.
        return self._load_identity_token_from_header(
            request,
        ) or self._load_identity_token_from_session(request)

    def process_request(self, request):
        assert hasattr(request, 'user'), (
            'The Fortmatic authentication middleware requires authentication '
            'middleware to be installed. Edit your MIDDLEWARE setting to insert '
            '`django.contrib.auth.middleware.AuthenticationMiddleware` before '
            '`fortmatic.django.auth.middleware.FortmaticAuthMiddleware`.'
        )

        identity_token = self._load_identity_token_from_sources(request)

        if not self._is_request_related_to_fortmatic(identity_token):
            return

        try:
            PhantomAuth().validate_token(identity_token)
        except (
            InvalidIdentityToken,
            InvalidSignerAddress,
            InvalidChallengeMessage,
        ):
            logout(request)
            return

        if request.user.is_anonymous:
            try:
                self._attempt_handling_anonymous_user(request, identity_token)
            except UnableToLoadUserFromIdentityToken:
                return

        if request.user.public_address != PhantomAuth().get_public_address(
            identity_token,
        ):
            logout(request)
            return

        if self.can_user_be_logged_in(request.user):
            self._persist_data_in_session(request, identity_token)
